from __future__ import division, print_function, absolute_import, unicode_literals

from unicodedata import east_asian_width
import six

__unicode_width_mapping = {'F': 2, 'H': 1, 'W': 2, 'Na': 1, 'A': 2, 'N': 1}


def is_unicode(s):
    return (six.PY2 and isinstance(s, unicode)) or (six.PY3 and isinstance(s, str))


def is_strlike(s):
    return isinstance(s, (six.string_types, bytes))


def unicode_width(s):
    if is_unicode(s):
        return sum(__unicode_width_mapping[east_asian_width(c)] for c in s)

    assert is_strlike(s), 's must be a string, not %s.' % type(s).__name__
    return len(s)


def to_unicode(s, encoding=None, errors='strict'):
    """
    Make unicode string from any value
    :param s:
    :param encoding:
    :param errors:
    :return: unicode
    """
    encoding = encoding or 'utf-8'

    if is_unicode(s):
        return s
    elif is_strlike(s):
        return s.decode(encoding, errors)
    else:
        if six.PY2:
            return str(s).decode(encoding, errors)
        else:
            return str(s)


def to_str(s, encoding=None, errors='strict'):
    """
    Make str from any value
    :param s:
    :param encoding:
    :param errors:
    :return: str (not unicode in Python2, nor bytes in Python3)
    """
    encoding = encoding or 'utf-8'

    if is_strlike(s):
        if six.PY2:
            return s.encode(encoding, errors) if isinstance(s, unicode) else s
        else:
            return s.decode(encoding, errors) if isinstance(s, bytes) else s
    else:
        return str(s)


def edge_just(left, right, width, fillchar=' '):
    padding = fillchar * max(1, width - unicode_width(left + right))
    return left + padding + right


def unicode_left(s, width):
    """Cut unicode string from left to fit a given width."""
    i = 0
    j = 0
    for ch in s:
        j += __unicode_width_mapping[east_asian_width(ch)]
        if width < j:
            break
        i += 1
    return s[:i]


def unicode_right(s, width):
    """Cut unicode string from right to fit a given width."""
    i = len(s)
    j = 0
    for ch in reversed(s):
        j += __unicode_width_mapping[east_asian_width(ch)]
        if width < j:
            break
        i -= 1
    return s[i:]



from __future__ import division, print_function, absolute_import, unicode_literals

import six


def get_single_item(d):
    """Get an item from a dict which contains just one item."""
    assert len(d) == 1, 'Single-item dict must have just one item, not %d.' % len(d)
    return next(six.iteritems(d))


def get_single_key(d):
    """Get a key from a dict which contains just one item."""
    assert len(d) == 1, 'Single-item dict must have just one item, not %d.' % len(d)
    return next(six.iterkeys(d))


def get_single_value(d):
    """Get a value from a dict which contains just one item."""
    assert len(d) == 1, 'Single-item dict must have just one item, not %d.' % len(d)
    return next(six.itervalues(d))



from __future__ import division, print_function, absolute_import, unicode_literals

import os
import sys
import locale
import re
import yaml
import six
from six.moves.urllib.request import urlopen
from jinja2 import Environment

from easy_menu.setting import arg_parser
from easy_menu.exceptions import SettingError, ConfigError, EncodingError
from easy_menu.util import CaseClass, cmd_util, string_util, term_util
from easy_menu.util.collection_util import get_single_item


DEFAULT_CONFIG_NAME = os.environ.get('EASY_MENU_CONFIG', 'easy-menu.yml')
URL_PATTERN = re.compile(r'^http[s]?://')


class Setting(CaseClass):
    """
    Manages all settings.
    """

    def __init__(self, config_path=None, work_dir=None, root_menu=None, encoding=None, lang=None, width=None,
                 stdin=None, stdout=None, stderr=None, cache=None):
        is_url = self._is_url(config_path)
        super(Setting, self).__init__(
            ('config_path', config_path),
            ('work_dir', self._search_work_dir(work_dir, config_path, is_url)),
            ('root_menu', {} if root_menu is None else root_menu),
            ('encoding', self._find_encoding(encoding, stdout)),
            ('lang', self._find_lang(lang)),
            ('width', width),
            ('stdin', stdin or sys.stdin),
            ('stdout', stdout or sys.stdout),
            ('stderr', stderr or sys.stderr),
            ('cache', {} if cache is None else cache)
        )

    def copy(self, **args):
        d = self.values()
        d.update(args)
        return Setting(**d)

    @staticmethod
    def _find_lang(lang):
        if not lang:
            # environment LANG is the first priority
            lang = os.environ.get('LANG')
        if not lang:
            lang = locale.getdefaultlocale()[0]
        return lang

    @staticmethod
    def _find_encoding(encoding, output):
        if not encoding:
            if hasattr(output, 'encoding'):
                encoding = output.encoding
        if not encoding:
            encoding = locale.getpreferredencoding()
        return encoding

    @staticmethod
    def _is_url(path):
        return path is not None and bool(URL_PATTERN.match(path))

    @staticmethod
    def _search_work_dir(work_dir, config_path, is_url):
        if work_dir is None:
            if config_path is not None:
                if not is_url:
                    return os.path.dirname(config_path)
        return work_dir

    def parse_args(self, argv):
        option, args = arg_parser.parser.parse_args(argv[1:])
        path = None

        if not args:
            pass
        elif len(args) == 1:
            path = args[0]
            if not self._is_url(path):
                path = os.path.abspath(path)
        else:
            arg_parser.parser.print_help()
            arg_parser.parser.exit(2)

        return self.copy(config_path=path, work_dir=option.work_dir, encoding=option.encoding, lang=option.lang,
                         width=option.width)

    def lookup_config(self):
        if self.config_path is None:
            d = os.path.abspath(self.work_dir) if self.work_dir else os.getcwd()

            while True:
                path = os.path.join(d, DEFAULT_CONFIG_NAME)
                if os.path.exists(path):
                    return self.copy(config_path=path)
                nd = os.path.dirname(d)
                if d == nd:
                    break
                d = nd
        return self

    def _load_data(self, is_command, path_or_url_or_cmdline):
        """
        Load data from one file, url or command line, then store to a dict.

        :param is_command: True if using command line output
        :param path_or_url_or_cmdline:
        :return: dict representation of data
        """

        if path_or_url_or_cmdline is None:
            raise SettingError('Not found configuration file.')

        # normalize file path
        is_url = self._is_url(path_or_url_or_cmdline)
        if not is_command and not is_url:
            if self.work_dir is not None and not os.path.isabs(path_or_url_or_cmdline):
                path_or_url_or_cmdline = os.path.join(self.work_dir, path_or_url_or_cmdline)

        # if already loaded, use cache data
        c = self.cache.get((is_command, path_or_url_or_cmdline))
        if c is not None:
            return c

        try:
            if is_command:
                # execute command
                term_util.universal_print(self.stdout, 'Executing: %s\n' % path_or_url_or_cmdline, self.encoding)

                # ignore return code and stderr
                data = cmd_util.capture_command(path_or_url_or_cmdline, self.work_dir, encoding=self.encoding)[1]
            elif self._is_url(path_or_url_or_cmdline):
                # read from URL
                term_util.universal_print(self.stdout, 'Reading from URL: %s\n' % path_or_url_or_cmdline, self.encoding)
                data = urlopen(path_or_url_or_cmdline).read()
            else:
                # read from file as bytes
                term_util.universal_print(self.stdout, 'Reading file: %s\n' % path_or_url_or_cmdline, self.encoding)
                with open(path_or_url_or_cmdline, 'rb') as f:
                    data = f.read()

            # If --encode option is not set, we use utf-8 for parsing YAML file.
            encoding = self.encoding or 'utf-8'
            data_str = data.decode(encoding)

            # apply jinja2 template rendering
            try:
                data_str = Environment().from_string(data_str).render()
            finally:
                pass

            # load as YAML string
            menu = yaml.load(data_str)

            # update cache data (Note: cache property is mutable!)
            self.cache[(is_command, path_or_url_or_cmdline)] = menu
        except IOError:
            raise ConfigError(path_or_url_or_cmdline, 'Failed to open.')
        except UnicodeDecodeError:
            raise EncodingError('Failed to decode with %s: %s' % (encoding, path_or_url_or_cmdline))
        except yaml.YAMLError as e:
            raise ConfigError(path_or_url_or_cmdline, 'YAML format error: %s' % string_util.to_unicode(str(e)))
        return menu

    def load_meta(self):
        root_menu = self._load_data(False, self.config_path)

        # overwrite working directory
        work_dir = root_menu.get('meta', {}).get('work_dir')
        return self.copy(work_dir=work_dir)

    def load_config(self):
        """
        Load the configuration file or url.

        If it contains 'include' sections, load them recursively.
        :return: updated Setting instance
        """

        def build_config(config, is_root, depth):
            """
            Validate and evaluate configuration recursively by using depth-first search.

            :param config: dict of the node or leaf
            :param is_root: True when the leaf is not allowed
            :param depth: indicator for the nesting level
            :return: validated dict of the node or leaf
            """
            # avoid for inclusion loops and stack overflow
            if depth >= 50:
                raise ConfigError(self.config_path, 'Nesting level too deep.')

            # config should be a dict which contains only one item except 'meta'
            if not isinstance(config, dict):
                raise ConfigError(self.config_path, 'Menu must be dict, not %s.' % type(config).__name__)

            config = dict((k, v) for k, v in config.items() if k not in ['meta'])

            if len(config) != 1:
                raise ConfigError(self.config_path, 'Menu should have only one item, not %s.' % len(config))

            name, content = get_single_item(config)
            t = type(content).__name__

            # encode key and leaf value
            name = string_util.to_unicode(name)
            is_leaf = isinstance(content, six.string_types)
            if is_leaf:
                content = string_util.to_unicode(content, self.encoding)

            if name == 'include':  # should be a leaf
                if not is_leaf:
                    raise ConfigError(self.config_path, '"include" section must have string content, not %s.' % t)
                return build_config(self._load_data(False, content), True, depth + 1)
            elif name == 'eval':  # should be a leaf
                if not is_leaf:
                    raise ConfigError(self.config_path, '"eval" section must have string content, not %s.' % t)
                return build_config(self._load_data(True, content), True, depth + 1)
            else:
                if isinstance(content, list):
                    if all(isinstance(child, six.string_types) for child in content):
                        content = [string_util.to_unicode(cmd, self.encoding) for cmd in content]
                    else:
                        content = [build_config(child, False, depth + 1) for child in content]
                elif is_leaf:
                    if is_root:
                        raise ConfigError(self.config_path, 'Root content must be list, not %s.' % t)
                else:
                    raise ConfigError(self.config_path, 'Content must be string or list, not %s.' % t)
            return {string_util.to_unicode(name, self.encoding): content}

        root_menu = build_config(self._load_data(False, self.config_path), True, 0)

        return self.copy(root_menu=root_menu)



from __future__ import division, print_function, absolute_import, unicode_literals


class EasyMenuError(Exception):
    """Base class of application specific exceptions"""


class EncodingError(EasyMenuError):
    """Encode error."""


class SettingError(EasyMenuError):
    """Setting error."""


class ConfigError(EasyMenuError):
    """Configuration error."""

    def __init__(self, path, msg=''):
        EasyMenuError.__init__(self, '%s: %s' % (path, msg))



from __future__ import division, print_function, absolute_import, unicode_literals

import sys
import six

from easy_menu.util import string_util, term_util
from easy_menu.util.collection_util import get_single_item, get_single_key, get_single_value
from easy_menu.exceptions import EncodingError, SettingError
from easy_menu.view import i18n

DEFAULT_WINDOW_WIDTH = 78
DEFAULT_PAGE_SIZE = 9


class Terminal(object):
    def __init__(self, root_menu, host, user, executor, width=None, page_size=None, _input=sys.stdin,
                 _output=sys.stdout, encoding=None, lang=None):
        """
        :param root_menu: dict of root menu
        :param host: host name string
        :param user: user name string
        :param executor: Executor instance
        :param width:
        :param page_size:
        :param _input:
        :param _output:
        :param encoding:
        :param lang: language setting
        :return:
        """
        # fields
        self.root_menu = root_menu
        self.host = host
        self.user = user
        self.executor = executor
        self.width = DEFAULT_WINDOW_WIDTH if width is None else width
        self.page_size = DEFAULT_PAGE_SIZE if page_size is None else page_size
        self._input = _input
        self._output = _output
        self.encoding = encoding
        self.lang = lang
        self.i18n = self._find_i18n(lang)

        if self.width < 40:
            raise SettingError('width must be equal or greater than 40: width=%s' % self.width)
        if not 0 < self.page_size <= 9:
            raise SettingError('page_size must be positive and one digit: page_size=%s' % self.page_size)

    @staticmethod
    def _find_i18n(lang):
        if not lang:
            return i18n.messages_en
        elif lang.lower().startswith('ja'):
            return i18n.messages_ja
        else:
            return i18n.messages_en

    def _num_pages(self, num_items):
        return max(1, (num_items + self.page_size - 1) // self.page_size)

    #
    # Design parts
    #
    def thin_line(self):
        return '-' * self.width

    def thick_line(self):
        return '=' * self.width

    def header_line(self):
        return string_util.edge_just(self.i18n.MSG_HOST % self.host, self.i18n.MSG_USER % self.user, self.width)

    @staticmethod
    def title_line(title):
        return ' ' * 2 + title

    def menu_line(self):
        return '+'.rjust(7, '-').ljust(self.width, '-')

    def pager_line(self, offset, num_pages):
        left = '<= [P]' if 0 < offset else '      '
        right = '[N] =>' if offset + 1 < num_pages else '      '
        middle = ('Page %d / %d' % (offset + 1, num_pages)).center(self.width - len(left) - len(right))
        return left + middle + right

    #
    # Building output string
    #
    def _get_header(self, title):
        return [
            self.header_line(),
            self.thick_line(),
            self.title_line(title),
            self.thin_line(),
        ]

    def _get_footer(self, message):
        return [
            self.thick_line(),
            message,
        ]

    def _get_description(self, item):
        d, c = get_single_item(item)
        return self.i18n.MSG_SUB_MENU % d if isinstance(c, list) else d

    def get_page(self, titles, page_items, offset, num_pages):
        """Make menu page string."""

        assert len(page_items) <= self.page_size, 'Number of page items must less or equal than page size.'

        title = string_util.unicode_right(' > '.join(titles), self.width - 4)

        pager_lines = [] if num_pages <= 1 else [
            self.pager_line(offset, num_pages),
            self.thin_line(),
        ]

        item_lines = [self.i18n.MSG_ITEM % (i + 1, self._get_description(item)) for i, item in enumerate(page_items)]

        quit_lines = [
            self.menu_line(),
            self.i18n.MSG_ITEM % (
                0, self.i18n.MSG_QUIT if len(titles) == 1 else self.i18n.MSG_RETURN % titles[-2]),
        ]

        message = self.i18n.MSG_INPUT_NUM % (0, len(page_items))
        return '\n'.join(self._get_header(title) + pager_lines + item_lines + quit_lines + self._get_footer(message))

    def get_confirm(self, description):
        """
        Make confirmation page string

        :param description: description for the command to execute
        :return: string
        """

        item_lines = [
            self.i18n.MSG_CONFIRM % description
        ]
        return '\n'.join(
            self._get_header(self.i18n.MSG_CONFIRM_TITLE) + item_lines +
            self._get_footer(self.i18n.MSG_CONFIRM_QUESTION))

    def get_before_execute(self, description):
        return '\n'.join(self._get_header(self.i18n.MSG_RUN_TITLE % description) + [''])

    def get_after_execute(self, return_code):
        result_lines = [
            self.thin_line(),
            'Return code: %d' % return_code,
        ]
        return '\n'.join(result_lines + self._get_footer(self.i18n.MSG_INPUT_ANY))

    #
    # Wait for input
    #
    def wait_input_char(self):
        ch = term_util.getch(self._input)
        if ch in ['\x03', '\x04']:
            # pressed C-c or C-d
            raise KeyboardInterrupt
        return ch

    #
    # Output
    #
    def _print(self, unicode_text):
        assert string_util.is_unicode(unicode_text), 'Text must be unicode: %s' % unicode_text

        try:
            term_util.universal_print(self._output, unicode_text, self.encoding)
        except (LookupError, UnicodeError):
            raise EncodingError('Failed to print menu: lang=%s, encoding=%s' % (self.lang, self.encoding))

    def _draw(self, unicode_text):
        term_util.clear_screen(self._input, self._output)
        self._print(unicode_text)

    #
    # Main loop
    #
    def wait_input_menu(self, menu_items, offset, num_pages):
        """
        :param menu_items: list of items
        :param offset: page offset
        :return: function(stack, offset => stack, offset)
        """
        num_items = len(menu_items)

        while True:
            ch = self.wait_input_char().lower()
            if ch == '0':
                return lambda s, o: (s[:-1], 0)
            elif ch.isdigit():
                index = offset * self.page_size + int(ch) - 1
                if index < num_items:
                    d = menu_items[index]
                    description, command = get_single_item(d)

                    # check if it is a sub menu
                    if isinstance(command, list):
                        # check if it is a command list
                        if not all(isinstance(child, six.string_types) for child in command):
                            return lambda s, o: (s + [d], 0)

                    def f(s, o):  # side effect only
                        self.execute_command(description, command)
                        return s, o

                    return f
            elif ch == 'n' and offset + 1 < num_pages:
                return lambda s, o: (s, o + 1)
            elif ch == 'p' and 0 < offset:
                return lambda s, o: (s, o - 1)

    def execute_command(self, description, command):
        """
        Confirm with prompt before executing command.

        :param description: description string
        :param command: command line string or list of command line string
        :return: None
        """

        # confirmation
        self._draw(self.get_confirm(description))

        while True:
            ch = self.wait_input_char().lower()
            if ch == 'y':
                break
            if ch == 'n' or ch == '\r':
                return

        # run command
        self._draw(self.get_before_execute(description))

        # if command is string, convert to a list
        if isinstance(command, six.string_types):
            command = [command]

        return_code = 0
        for cmd in command:
            return_code = self.executor.execute(
                cmd, stdin=self._input, stdout=self._output, stderr=self._output, encoding=self.encoding)

            if return_code == 130:
                # maybe interrupted
                self._print('\n')
            # if a command fails, the successors will not run
            if return_code != 0:
                break

        self._print(self.get_after_execute(return_code))
        self.wait_input_char()  # wait for any input

    def loop(self):
        stack = [self.root_menu]
        offset = 0  # current page index

        while stack:
            titles = [get_single_key(d) for d in stack]
            items = get_single_value(stack[-1])
            num_pages = self._num_pages(len(items))

            # apply offset
            page_items = items[self.page_size * offset:self.page_size * (offset + 1)]

            self._draw(self.get_page(titles, page_items, offset, num_pages))

            f = self.wait_input_menu(get_single_value(stack[-1]), offset, num_pages)
            stack, offset = f(stack, offset)

        term_util.clear_screen(self._input, self._output)



from __future__ import division, print_function, absolute_import, unicode_literals


class CaseClass(object):
    """
    Implementation like Scala's case class

    This class can order if and only if all the element can order.

    Example:
        class Coord(CaseClass):
            def __init__(self, x, y):
                super(Coord, self).__init__(('x', x), ('y', y))

        a = Coord(123, 45)
        a.x  # 123
        a.y  # 45
        str(a)  # 'Coord(x=123, y=45)'
        b = a.copy(y=54)
        str(b)  # 'Coord(x=123, y=54)'
        a < b  # True
    """

    def __init__(self, *args):
        """
        :param args: list of tuple of field key and value
        """
        keys = []
        for k, v in args:
            if k in keys:
                raise ValueError('Found duplicate key name: %s' % k)
            keys.append(k)
            setattr(self, k, v)
        self._keys = keys

    def __cmp__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError('unorderable types: %s() < %s()' % (self.__class__.__name__, other.__class__.__name__))

        for k in self._keys:
            a, b = getattr(self, k), getattr(other, k)
            if a is not None or b is not None:
                if a < b:
                    return -1
                if a > b:
                    return 1
        return 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        for k in self._keys:
            a, b = getattr(self, k), getattr(other, k)
            if a is not None or b is not None:
                if a != b:
                    return False
        return True

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, ', '.join('%s=%r' % (k, getattr(self, k)) for k in self._keys))

    def values(self):
        """
        :return: key-value dict : { string: any }
        """
        return dict((k, getattr(self, k)) for k in self._keys)



from __future__ import division, print_function, absolute_import, unicode_literals

import sys
import os
import time
import subprocess

if os.name == 'nt':
    # for Windows
    import msvcrt
else:
    # for Unix/Linux/Mac
    import termios
    import tty

from easy_menu.util import string_util

LAST_GETCH_TIME = 0.0
LAST_GETCH_CHAR = ''


def _wrap_termios(_input, func):
    assert hasattr(_input, 'fileno'), 'Invalid input device.'

    fd = _input.fileno()
    old_settings = None
    try:
        try:
            old_settings = termios.tcgetattr(fd)
            tty.setraw(fd)
        except termios.error:
            pass
        ret = func()
    finally:
        if old_settings:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

    return ret


def getch(_input=sys.stdin):
    """Wait and get one character from input"""

    global LAST_GETCH_TIME
    global LAST_GETCH_CHAR

    if os.name == 'nt':
        ch = msvcrt.getch()
    else:
        ch = _wrap_termios(_input, lambda: _input.read(1))

    try:
        uch = string_util.to_unicode(ch, 'ascii')
    except UnicodeError:
        return ''

    t = time.time()

    # check key repeat
    if LAST_GETCH_CHAR == uch:
        if t < LAST_GETCH_TIME + 0.3:
            LAST_GETCH_TIME = t
            return ''

    LAST_GETCH_TIME = t
    LAST_GETCH_CHAR = uch

    return uch


def clear_screen(_input=sys.stdin, _output=sys.stdout):
    """Clear terminal screen."""

    if not _output.isatty():
        return

    cmd = 'cls' if os.name == 'nt' else 'clear'
    subprocess.call(cmd, shell=True, stdin=_input, stdout=_output, stderr=_output)


def universal_print(output, str_or_bytes, encoding='utf-8'):
    """Print unicode or bytes universally"""

    if string_util.is_unicode(str_or_bytes):
        bs = str_or_bytes.encode(encoding)
    else:
        bs = str_or_bytes

    if hasattr(output, 'buffer'):
        output.buffer.write(bs)
    else:
        try:
            output.write(bs)
        except TypeError:
            output.write(str_or_bytes)

    output.flush()



from __future__ import division, print_function, absolute_import, unicode_literals

import sys
import subprocess
from easy_menu.util import string_util

# Note: workaround for http://bugs.python.org/issue8513
NOT_USE_SHELL = sys.version_info[:2] == (3, 2) and not sys.platform == 'win32'


def execute_command(cmd, work_dir, stdin, stdout, stderr, encoding='utf-8'):
    """
    Execute external command

    :param cmd: command line string
    :param work_dir: working directory
    :param stdin: standard input
    :param stdout: standard output
    :param stderr: standard error
    :return: return code
    """
    assert string_util.is_unicode(cmd), 'cmd must be unicode string, not %s' % type(cmd).__name__

    try:
        if NOT_USE_SHELL:
            ret_code = subprocess.call(
                ['/bin/sh', '-c', cmd.encode(encoding)],
                shell=False,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                cwd=work_dir,
            )
        else:
            ret_code = subprocess.call(
                cmd.encode(encoding),
                shell=True,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                cwd=work_dir,
            )
    except KeyboardInterrupt:
        ret_code = 130

    return ret_code


def capture_command(cmd, work_dir, stdin=sys.stdin, encoding='utf-8'):
    """
    Execute external command and capture output

    :param cmd: command line string
    :param work_dir: working directory
    :param stdin: standard input
    :return: tuple of return code, stdout data and stderr data
    """
    assert string_util.is_unicode(cmd), 'cmd must be unicode string, not %s' % type(cmd).__name__

    stdout_data, stderr_data = None, None

    try:
        if NOT_USE_SHELL:
            p = subprocess.Popen(
                ['/bin/sh', '-c', cmd.encode(encoding)],
                shell=False,
                stdin=stdin,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=work_dir
            )
        else:
            p = subprocess.Popen(
                cmd.encode(encoding),
                shell=True,
                stdin=stdin,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=work_dir
            )
        stdout_data, stderr_data = p.communicate()
        ret_code = p.returncode
    except KeyboardInterrupt:
        ret_code = 130

    return ret_code, stdout_data, stderr_data



from __future__ import division, print_function, absolute_import, unicode_literals

import sys
from easy_menu.view import Terminal
from easy_menu.controller import CommandExecutor
from easy_menu.setting.setting import Setting
from easy_menu.logger import SystemLogger
from easy_menu.util import network_util
from easy_menu.exceptions import EasyMenuError


def main(stdin=None, stdout=None, stderr=None):
    """
    Main function
    """

    base_setting = Setting(stdin=stdin, stdout=stdout, stderr=stderr)

    try:
        setting = base_setting.parse_args(sys.argv).lookup_config().load_meta().load_config()
        executor = CommandExecutor(setting.work_dir, SystemLogger(setting.encoding))

        t = Terminal(
            setting.root_menu,
            network_util.get_hostname(),
            network_util.get_username(),
            executor,
            _input=setting.stdin,
            _output=setting.stdout,
            encoding=setting.encoding,
            lang=setting.lang,
            width=setting.width
        )

        t.loop()
    except KeyboardInterrupt:
        pass
    except EasyMenuError as e:
        base_setting.stdout.write('%s: %s\n' % (e.__class__.__name__, e))
        return 2
    return 0



from __future__ import division, print_function, absolute_import, unicode_literals

import os
import sys
try:
    import syslog
    SYSLOG_AVAILABLE = True
except ImportError:
    SYSLOG_AVAILABLE = False
from easy_menu.logger.logger import Logger
from easy_menu.util import string_util


class SystemLogger(Logger):
    def __init__(self, encoding='utf-8'):
        self.name = os.path.basename(sys.argv[0])
        self.encoding = encoding
        super(SystemLogger, self).__init__('SystemLogger[%s]' % self.name)

    def _log(self, priority, message):
        if SYSLOG_AVAILABLE:
            syslog.openlog(self.name, syslog.LOG_PID)
            syslog.syslog(priority, string_util.to_str(message, self.encoding))
            syslog.closelog()



from __future__ import division, print_function, absolute_import, unicode_literals

from abc import ABCMeta, abstractmethod
import six
from easy_menu.util import CaseClass

# copied from syslog module
LOG_ALERT = 1
LOG_AUTH = 32
LOG_CONS = 2
LOG_CRIT = 2
LOG_CRON = 72
LOG_DAEMON = 24
LOG_DEBUG = 7
LOG_EMERG = 0
LOG_ERR = 3
LOG_INFO = 6
LOG_KERN = 0
LOG_LOCAL0 = 128
LOG_LOCAL1 = 136
LOG_LOCAL2 = 144
LOG_LOCAL3 = 152
LOG_LOCAL4 = 160
LOG_LOCAL5 = 168
LOG_LOCAL6 = 176
LOG_LOCAL7 = 184
LOG_LPR = 48
LOG_MAIL = 16
LOG_NDELAY = 8
LOG_NEWS = 56
LOG_NOTICE = 5
LOG_NOWAIT = 16
LOG_PERROR = 32
LOG_PID = 1
LOG_SYSLOG = 40
LOG_USER = 8
LOG_UUCP = 64
LOG_WARNING = 4


@six.add_metaclass(ABCMeta)
class Logger(CaseClass):
    """abstract logger class"""

    def __init__(self, ident):
        """
        :param ident: identity for logger
        """
        CaseClass.__init__(self, ('ident', ident))

    def info(self, message):
        self._log_message(LOG_INFO, message)

    def warn(self, message):
        self._log_message(LOG_WARNING, message)

    def error(self, message):
        self._log_message(LOG_ERR, message)

    def traceback(self):
        """do nothing by default"""

    def _log_message(self, priority, message):
        prefix = {
            LOG_INFO: '[INFO] ',
            LOG_WARNING: '[WARN] ',
            LOG_ERR: '[ERROR]'
        }[priority]
        msg = '%s%s' % (prefix, message)
        self._log(priority, msg)

    @abstractmethod
    def _log(self, priority, message):
        """log raw message"""



# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import, unicode_literals

MSG_HOST = 'ホスト名: %s'
MSG_USER = '実行ユーザ: %s'
MSG_ITEM = '  [%d] | %s'
MSG_SUB_MENU = '%s へ'
MSG_QUIT = '終了'
MSG_RETURN = '%s に戻る'
MSG_INPUT_NUM = '番号を入力してください (%d-%d): '
MSG_INPUT_ANY = '何かキーを押すとメニューに戻ります...'
MSG_CONFIRM_TITLE = '実行確認'
MSG_CONFIRM = '  %s を行います。'
MSG_CONFIRM_QUESTION = 'よろしいですか? (y/n) [n]: '
MSG_RUN_TITLE = '実行: %s'



from __future__ import division, print_function, absolute_import, unicode_literals

import os
import socket
import getpass


def get_hostname():
    hostname = socket.gethostname()
    nickname = os.environ.get('NICKNAME')
    return hostname + (' (%s)' % nickname if nickname else '')


def get_username():
    return getpass.getuser()



from __future__ import division, print_function, absolute_import, unicode_literals

MSG_HOST = 'Host: %s'
MSG_USER = 'User: %s'
MSG_ITEM = '  [%d] | %s'
MSG_SUB_MENU = 'Go to %s'
MSG_QUIT = 'Quit'
MSG_RETURN = 'Return to %s'
MSG_INPUT_NUM = 'Press menu number (%d-%d): '
MSG_INPUT_ANY = 'Press any key to continue...'
MSG_CONFIRM_TITLE = 'Confirmation'
MSG_CONFIRM = '  Would execute: %s'
MSG_CONFIRM_QUESTION = 'Do you really want to execute? (y/n) [n]: '
MSG_RUN_TITLE = 'Executing: %s'



from __future__ import division, print_function, absolute_import, unicode_literals

from optparse import OptionParser

VERSION = 'easy-menu %s' % __import__('easy_menu').__version__

USAGE = """%prog [options...] [<config_path> | <config_url>]"""


def __get_parser():
    p = OptionParser(version=VERSION, usage=USAGE)

    p.add_option(
        '--encode', dest='encoding', default=None, type='string', metavar='ENCODING',
        help='set output encoding to ENCODING'
    )

    p.add_option(
        '--lang', dest='lang', default=None, type='string', metavar='LANG',
        help='set language to LANG (in RFC 1766 format)'
    )

    p.add_option(
        '--width', dest='width', default=None, type='int', metavar='WIDTH',
        help='set window width to WIDTH'
    )

    p.add_option(
        '-d', '--work-dir', dest='work_dir', default=None, type='string', metavar='DIR',
        help='set working directory to DIR'
    )

    return p


parser = __get_parser()



from __future__ import division, print_function, absolute_import, unicode_literals

from easy_menu.util import cmd_util


class CommandExecutor(object):
    def __init__(self, work_dir=None, logger=None):
        self.work_dir = work_dir
        self.logger = logger

    def execute(self, cmd, stdin, stdout, stderr, encoding):
        self.logger.info('Command started: %s' % cmd)
        ret_code = cmd_util.execute_command(cmd, self.work_dir, stdin, stdout, stderr, encoding)
        self.logger.info('Command ended with return code: %d' % ret_code)
        return ret_code
